
/****************************************************************************

  Demos Tool

	Engine code, independant from OS interface.

	Cyril Gantin (cgg@planetquake.com)

 ****************************************************************************/



#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <io.h>
#include <string.h>

#define MAX_PATH 256
#define DEMO_EXT ".dm_68"
#define DEMO_PATH "demos\\"

class demoinfo_c {
	public:
		bool	valid;
		bool	cpm;
		bool	timeIsSet;
		char	timeString[16];
		char	indexString[8];
		char	mapName[128];
		char	playerName[128];
		char	playerCountry[128];

		void SetFromFilename(char *);
		int GetTime();
		void output();

	private:
		int		time;
};



//global vars
char gPlayerName[MAX_PATH];		//settings
int gNumDemosPerMap;			//settings
int gStatsFilesNum,gStatsSize;	//stats tracking

//from winmain.cpp
void trap_print(char *s);
void trap_error(char *s);


// ==============================================
//
//   General functions
//
// ==============================================

char *va(char *fmt, ...) {
	va_list argptr;
	static char buf[1024];

	va_start (argptr,fmt);
	vsprintf (buf,fmt,argptr);
	va_end (argptr);

	return buf;
}

//not the safest way to test for a file to exist, but efficient enough for this app
bool fileExists(char *filename) {
	FILE *file;
	file = fopen(filename,"r");
	if (!file)
		return false;
	fclose(file);
	return true;
}

//Parse a NULL-terminated string from cursor c to retrieve the next argument, copied 
//into the buffer passed through the arg pointer.
//The cursor c is then set to the next argument first char
void getNextArg(char **c,char *arg,int len,char separator) {
	int i = 0;

	//move cursor to the first argument char
	while (**c == separator)
		(*c)++;

	//copy argument to buffer
	while (**c && **c != separator && i < len) {
		*(arg+i) = **c;
		(*c)++; i++;
	}
	if (i == len) {		//overflow, make the output string (arg) NULL-terminated and move cursor to the end of the argument
		*(arg+i-1) = 0;
		while (**c != separator && **c)	//the input string passed through **c MUST be NULL-terminated,
			(*c)++;
	} else {
		*(arg+i) = 0;	//terminate the argument string
	}

	//move cursor to next argument first char
	while (**c == separator)
		(*c)++;
}

bool isTimeString(char *c) {
	int i,len = strlen(c);

	if (len < 5 || len > 9)
		return false;

	if (!strcmp(c,"TimeHere")) {
		return true;
	}

	i = 0;
	while (*c) {
		if (len - i == 4 || len - i == 7) {
			if (*c != '.')
				return false;
		} else if (*c < '0' || *c > '9') {
			return false;
		}
		c++; i++;
	}

	return true;
}


//
// demoinfo_c
//

//used for debugging purpose
void demoinfo_c::output() {
	if (!valid) {
		trap_print("Invalid demoinfo.\r\n");
		return;
	}
	trap_print(va("TimeString : %s\r\n",timeString));
	trap_print(va("TimeIsSet  : %s\r\n",(timeIsSet)?"Yes":"No"));
	trap_print(va("Time       : %i\r\n",time));
	trap_print(va("CPM        : %s\r\n",(cpm)?"Yes":"No"));
	trap_print(va("Map        : %s\r\n",mapName));
	trap_print(va("Player     : %s\r\n",playerName));
	trap_print(va("Country    : %s\r\n",playerCountry));
	trap_print(va("Index      : %s\r\n",indexString));
}

// returns demo time (-1 if time can't be set)
int demoinfo_c::GetTime() {
	char arg[3][16];
	char *c;
	int argn;
	int i;

	if (time)	//time already set
		return time;

	//time can't be set
	if (!valid || !timeIsSet) {
		time = -1;
		return time;
	}

	//set time from timeString
	c = timeString;
	argn = 0;
	i = 0;

	while (*c && argn<3) {
		if (*c == '.') {	//terminate current arg with NULL and move to the next arg
			arg[argn][i] = 0;
			argn++;
			i = 0;
		} else {
			arg[argn][i] = *c;
			i++;
		}
		c++;
	}

	if (argn == 2) {
		time = atoi(arg[0]) *60000;
		time += atoi(arg[1]) *1000;
		time += atoi(arg[2]);
	} else if (argn == 1) {
		time = atoi(arg[0]) *1000;
		time += atoi(arg[1]);
	} else {
		time = -1;
	}

	return time;
}


//parse a demo filename
//demoinfo_c.valid will be set to false if the filename is not a valid Defrag demo name
void demoinfo_c::SetFromFilename(char *stringOrigin) {
	char arg[128];
	char *mapnameEndPos,*c;
	bool latchedCpm;
	int i;

	c = stringOrigin;

	//set time string and cpm flag
	arg[0] = 0;
	do {
		latchedCpm = (strcmp(arg,"cpm")) ? false : true;
		mapnameEndPos = c;
		getNextArg(&c,arg,sizeof(arg),'_');
	} while (!isTimeString(arg) && *c);

	if (*c == 0) {
		valid = false;
		return;
	}

	strcpy(timeString,arg);
	timeIsSet = (strcmp(arg,"TimeHere")) ? true : false;
	cpm = latchedCpm;

	//set mapname
	if (latchedCpm) {
		mapnameEndPos -= 5;	//get rid of the cpm tag: _cpm_
	} else {
		mapnameEndPos--;	//get rid of the separator: _
	}
	if (mapnameEndPos - stringOrigin > 127 || mapnameEndPos - stringOrigin < 1) {	//abnormal argument length
		valid = false;
		return;
	}
	strncpy(mapName,stringOrigin,mapnameEndPos-stringOrigin);
	mapName[mapnameEndPos-stringOrigin] = 0;

	//set player name
	//since it should not contain any '_', don't perform any error handling
	getNextArg(&c,playerName,sizeof(playerName),'_');

	if (*c == 0) {	//unexpected end of string
		valid = false;
		return;
	}

	//set player country
	//should not contain any '_'
	i = 0;
	while(*c && *c != '.' && *c != '_') {
		playerCountry[i] = *c;
		c++; i++;
	}
	playerCountry[i] = 0;

	//set index string if there
	if (*c && *c == '_' && !timeIsSet) {
		c++;
		for (i=0;i<3 && *c;i++) {
			indexString[i] = *c;
			c++;
		}
		indexString[3] = 0;
	}

	//only demo extention should remain
	if (*c == 0 || strcmp(c,DEMO_EXT)) {
		valid = false;
		return;
	}

	//filename string parsed, and passed through all error tests, we are done
	valid = true;
	time = 0;	//time must be set from GetTime()
}



// ==============================================
//
//   Init
//
// ==============================================

//GetNameFromQ3CFG() attempts to catch df_name value from q3config.cfg
void GetNameFromQ3CFG() {
	char buf[1024*32];
	char line[256];
	char arg[256];
	char *c,*c2;
	FILE *cfg;
	int i;

	gPlayerName[0] = 0;

	cfg = fopen("q3config.cfg","r");
	if (!cfg) {
		return;
	}
	i = fread(buf,sizeof(char),sizeof(buf),cfg);
	buf[i] = 0;
	fclose(cfg);

	c = buf;
	do {
		getNextArg(&c,line,sizeof(line),'\n');
		c2 = line;
		getNextArg(&c2,arg,sizeof(arg),' ');	//is probably "seta" or "bind"
		getNextArg(&c2,arg,sizeof(arg),' ');	//is probably Cvar name
		if (strcmp(arg,"df_name"))
			continue;
		//c2 cursor is located on df_name value
		if (*c2 == '"')
			c2++;
		i = 0;
		while (*c2) {
			if (*c2 != '"') {
				gPlayerName[i] = *c2;
				i++;
			}
			c2++;
		}
		return;
	} while (*c);
}

void DT_UpdateConfig() {
	FILE *config;
	char *s;

	config = fopen("_demostool.cfg","wb");
	if (!config)
		return;

	if (gPlayerName[0])
		s = va("%s %s\n%s %i","playername",gPlayerName,"nDemosPerMap",gNumDemosPerMap);
	else
		s = va("%s %i\n","nDemosPerMap",gNumDemosPerMap);
	fwrite(s,sizeof(char),strlen(s),config);
	fclose(config);
}

bool GetDTConfig() {
	char buf[1024];
	char line[128];
	char arg[128];
	char *c,*c2;
	FILE *config;
	int i;

	config = fopen("_demostool.cfg","rb");
	if (!config)
		return false;
	i = fread(buf,sizeof(char),sizeof(buf),config);
	buf[i] = 0;
	fclose(config);

	c = buf;
	while (*c) {
		getNextArg(&c,line,sizeof(line),'\n');
		c2 = line;
		getNextArg(&c2,arg,sizeof(arg),' ');

		if (!strcmp(arg,"playername")) {
			getNextArg(&c2,arg,sizeof(arg),' ');
			strncpy(gPlayerName,arg,sizeof(gPlayerName));
		} else if (!strcmp(arg,"nDemosPerMap")) {
			getNextArg(&c2,arg,sizeof(arg),' ');
			gNumDemosPerMap = atoi(arg);
		}
	}
	return true;
}

//must be called from OS interface
void DT_init() {
	if (!GetDTConfig()) {
		GetNameFromQ3CFG();
		gNumDemosPerMap = 2;
		DT_UpdateConfig();
	}
}


//must be called from OS interface
bool DT_CheckDefrag() {
	FILE *descript;
	char buf[256];
	char arg[256];
	char *c;
	int i,version;

	//check if we are running from Defrag directory
	descript = fopen("description.txt","rb");
	if (!descript) {
		trap_error("Please run from Defrag directory");
		return false;
	}
	i = fread(buf,sizeof(char),sizeof(buf),descript);
	buf[i] = 0;
	fclose(descript);

	c = buf;
	getNextArg(&c,arg,sizeof(arg),' ');
	if (strcmp(arg,"Defrag")) {
		trap_error("Please run from Defrag directory");
		return false;
	}

	//check Defrag version (must be superior to 1.61)
	i = 0; version = 0;
	while (*c && i < 4) {
		if (*c >= '0' && *c <= '9') {
			if (i == 0)
				version += (int)(*c - '0') *100;
			else if (i == 2)
				version += (int)(*c - '0') *10;
			else if (i == 3)
				version += (int)(*c - '0');
		}
		c++; i++;
	}

	if (version > 161)
		return true;

	trap_error("This application requires Defrag 1.62 or above");
	return false;
}




// ==============================================
//
//   Pre-steps
//
// ==============================================

//rename old .dm_66 extensions to .dm_68
void ConvertDM66Demos() {
	char newname[MAX_PATH];
	char oldname[MAX_PATH];
	struct _finddata_t c_file;
	long hFile;
	int NoMoreFile;

	gStatsFilesNum = 0;
	NoMoreFile = 0;
	hFile = _findfirst(va("%s*.dm_66",DEMO_PATH),&c_file);

	while (hFile != -1L && !NoMoreFile) {
		strcpy(oldname,DEMO_PATH);
		strcat(oldname,c_file.name);
		strcpy(newname,DEMO_PATH);
		strcat(newname,c_file.name);
		newname[MAX_PATH-1] = 0;	//make sure the string is NULL-terminated
		newname[strlen(newname)-1] = '8';
		if (rename(oldname,newname)) {
			trap_print(va("warning: Can't rename %s.\r\n",c_file.name));
			NoMoreFile = _findnext( hFile, &c_file );	//will set NoMoreFile to 0 if a next file is found
		} else {
			_findclose(hFile);	//must release the search-handle and restart search after any file operation
			hFile = _findfirst(va("%s*.dm_66",DEMO_PATH),&c_file);
			gStatsFilesNum++;
		}
	}
	_findclose(hFile);

	if (gStatsFilesNum)
		trap_print(va("found and converted %i dm_66 demos.\r\n\r\n",gStatsFilesNum));
}

//rename .dm_67 extensions to .dm_68
void ConvertDM67Demos() {
	char newname[MAX_PATH];
	char oldname[MAX_PATH];
	struct _finddata_t c_file;
	long hFile;
	int NoMoreFile;

	gStatsFilesNum = 0;
	NoMoreFile = 0;
	hFile = _findfirst(va("%s*.dm_67",DEMO_PATH),&c_file);

	while (hFile != -1L && !NoMoreFile) {
		strcpy(oldname,DEMO_PATH);
		strcat(oldname,c_file.name);
		strcpy(newname,DEMO_PATH);
		strcat(newname,c_file.name);
		newname[MAX_PATH-1] = 0;	//make sure the string is NULL-terminated
		newname[strlen(newname)-1] = '8';
		if (rename(oldname,newname)) {
			trap_print(va("warning: Can't rename %s.\r\n",c_file.name));
			NoMoreFile = _findnext( hFile, &c_file );	//will set NoMoreFile to 0 if a next file is found
		} else {
			_findclose(hFile);	//must release the search-handle and restart search after any file operation
			hFile = _findfirst(va("%s*.dm_67",DEMO_PATH),&c_file);
			gStatsFilesNum++;
		}
	}
	_findclose(hFile);

	if (gStatsFilesNum)
		trap_print(va("found and converted %i dm_67 demos.\r\n\r\n",gStatsFilesNum));
}


//make any mapname lowercase
void SanitizeFilenames() {
	struct _finddata_t c_file;
	long hFile;
	demoinfo_c demoinfo;
	char newname[MAX_PATH];
	char oldname[MAX_PATH];
	char *c;
	int NoMoreFile;
	bool needRename;

	hFile = _findfirst(va("%s*%s",DEMO_PATH,DEMO_EXT),&c_file);
	NoMoreFile = 0;

	while (hFile != -1L && !NoMoreFile) {
		demoinfo.SetFromFilename(c_file.name);
		needRename = false;

		//not a valid defrag demo
		if (!demoinfo.valid) {
			NoMoreFile = _findnext(hFile,&c_file);	//will set NoMoreFile to 0 if a next file is found
			continue;
		}

		//check the mapname, should be lower-case only.
		c = demoinfo.mapName;
		while (*c) {
			if (*c >= 'A' && *c <= 'Z') {
				*c |= (char)(1<<5);
				needRename = true;
			}
			c++;
		}

		if (!needRename) {
			NoMoreFile = _findnext(hFile,&c_file);
			continue;
		}

		strcpy(oldname,DEMO_PATH);
		strcat(oldname,c_file.name);
		strcpy(newname,va("%s%s_%s%s_%s_%s%s%s",
			DEMO_PATH,demoinfo.mapName,(demoinfo.cpm)?"cpm_":"",demoinfo.timeString,
			demoinfo.playerName,demoinfo.playerCountry,(demoinfo.timeIsSet)?"":demoinfo.indexString,DEMO_EXT));

		newname[MAX_PATH-1] = 0;	//make sure the string is NULL-terminated
		if (rename(oldname,newname)) {
			NoMoreFile = _findnext(hFile,&c_file );	//rename() failed, move on to next file
		} else {
			_findclose(hFile);	//must release the search-handle and restart search after any file operation
			hFile = _findfirst(va("%s*%s",DEMO_PATH,DEMO_EXT),&c_file);
			gStatsFilesNum++;
		}
	}
	_findclose(hFile);

}




// ==============================================
//
//   Step 1: Rename demos from .dem files
//
// ==============================================


void PerformDemFile(char *filepath,char *filename) {
	FILE *dem;
	char buf[1024*64];
	char oldname[MAX_PATH];
	char newname[MAX_PATH];
	char mapname[128];
	char playername[128];
	char country[128];
	char time[16];
	char index[4];
	char *c;
	bool cpm;
	int i;

	//copy file content to buffer
	dem = fopen(filepath,"r");
	if (!dem) {
		return;
	}
	i = fread(buf,sizeof(char),sizeof(buf),dem);
	buf[i] = 0;
	fclose(dem);

	//get mapname from .dem file name
	strcpy(mapname,filename);
	mapname[strlen(filename)-4] = 0;

	//parse buffer
	c = buf;
	while (*c) {
		//get demo info
		getNextArg(&c,index,sizeof(index),' ');
		getNextArg(&c,time,sizeof(time),' ');
		if (!strcmp (time,"cpm")) {
			cpm = true;
			getNextArg(&c,time,sizeof(time),' ');
		} else {
			cpm = false;
		}
		getNextArg(&c,playername,sizeof(playername),' ');
		getNextArg(&c,country,sizeof(country),' ');

		//replace each ':' from time string with '.'
		for (i=0;i<(int)strlen(time);i++)
			if (time[i] == ':')
				time[i] = '.';

		//errors handling: a few ways to check the validity of data contained in .dem files
		if (*c == 0 || !isTimeString(time) || index[3] != 0) {
			trap_print(va("error: %s is invalid.\r\n",filename));
			return;
		}
		for (i=0;i<3;i++) {
			if (index[i] < '0' || index[i] > '9') {
				trap_print(va("error: %s is invalid.\r\n",filename));
				return;
			}
		}
		if (*c != '/' && *(c+1) != ' ') {
			trap_print(va("error: %s is invalid.\r\n",filename));
			return;
		}
		if (*(c+1))
			c += 2;	//move to first argument of the next demo
		
		//rename the demo
		strncpy(oldname,va("%s%s_%sTimeHere_%s_%s_%s%s",DEMO_PATH,mapname,(cpm)?"cpm_":"",playername,country,index,DEMO_EXT),MAX_PATH);
		strncpy(newname,va("%s%s_%s%s_%s_%s%s",DEMO_PATH,mapname,(cpm)?"cpm_":"",time,playername,country,DEMO_EXT),MAX_PATH);

		if (!fileExists(newname)) {	//if destination filename already exists, just ignore and move on
			if (rename(oldname,newname)) {
				trap_print(va("warning: can't rename %s.\r\n",oldname));
			} else {
				gStatsFilesNum++;
			}
		}
	}
}

void RenameRegisteredDemos() {
	struct _finddata_t c_file;
	long hFile;
	char path[MAX_PATH];
	int NoMoreFile;

	gStatsFilesNum = 0;
	NoMoreFile = 0;
	hFile = _findfirst(va("%s*.dem",DEMO_PATH),&c_file);

	while (hFile != -1L && !NoMoreFile) {
		strcpy(path,DEMO_PATH);
		strcat(path,c_file.name);
		PerformDemFile(path,c_file.name);
		if (remove(path)) {	//delete the .dem file we have just processed
			trap_print(va("warning: can't delete %s.\r\n",c_file.name));
			NoMoreFile = _findnext( hFile, &c_file );	//will set NoMoreFile to 0 if a next file is found
		} else {
			_findclose(hFile);	//must release the search-handle and restart search after any-file operation
			hFile = _findfirst(va("%s*.dem",DEMO_PATH),&c_file);
		}
	}
	_findclose(hFile);

	trap_print(va("%i demos renamed.\r\n",gStatsFilesNum));
}





// ==============================================
//
//   Step 2: Delete Demos whose time is not set
//
// ==============================================


void DeleteUnfinishedDemos() {
	struct _finddata_t c_file;
	long hFile;
	demoinfo_c demoinfo;
	int NoMoreFile;

	gStatsFilesNum = 0;
	gStatsSize = 0;

	NoMoreFile = 0;
	hFile = _findfirst(va("%s*%s",DEMO_PATH,DEMO_EXT),&c_file);

	while (hFile != -1L && !NoMoreFile) {
		demoinfo.SetFromFilename(c_file.name);

		if (!demoinfo.valid || demoinfo.timeIsSet) {
			NoMoreFile = _findnext( hFile, &c_file );
			continue;
		}

		//Demo's time-string is tagged TimeHere, delete it.
		if (remove(va("%s%s",DEMO_PATH,c_file.name))) {
			trap_print(va("Warning: Can't delete %s.\r\n",c_file.name));
			NoMoreFile = _findnext( hFile, &c_file );	//will set NoMoreFile to 0 if a next file is found
		} else {
			gStatsFilesNum++;
			gStatsSize += c_file.size / 1024;
			_findclose(hFile);	//must release the search-handle and restart search after any file operation
			hFile = _findfirst(va("%s*%s",DEMO_PATH,DEMO_EXT),&c_file);
		}
	}
	_findclose(hFile);

	trap_print(va("Deleted %i files. (%i kb)\r\n",gStatsFilesNum,gStatsSize));
}





// ==============================================
//
//   Step 3: Delete demos in excess
//
// ==============================================


//returns false if operation failed
bool RenameToTmp(char *filename) {
	char newname[MAX_PATH];

	strncpy(newname,filename,MAX_PATH);
	newname[strlen(filename)-strlen(DEMO_EXT)] = 0;
	strcat(newname,".tmp");
	if (rename(filename,newname))
		return false;
	else
		return true;
}

//returns false if operation failed
bool RenameTmpToDemoExt(char *filename) {
	char newname[MAX_PATH];

	strncpy(newname,filename,MAX_PATH);
	newname[strlen(filename)-4] = 0;
	strcat(newname,DEMO_EXT);
	if (rename(filename,newname))
		return false;
	else
		return true;
}


//process vq3 and cpm demos separately
void ProcessDemosPerMap(char *mapname,bool processCPM) {
	struct _finddata_t c_file;
	long hFile;
	demoinfo_c demoinfo;
	int NoMoreFile,FileOpFailed;
	int besttime[16];
	int besttime_n;
	int time,maxtime,i,j;
	int nDemosKept;

	// First pass: get the maximum time value from which any above demo will be deleted
	besttime_n = 0;
	hFile = _findfirst(va("%s%s%s*_%s_*%s",DEMO_PATH,mapname,(processCPM)?"_cpm_":"",gPlayerName,DEMO_EXT),&c_file);
	if (hFile != -1L) do {
		//get time from demo filename
		demoinfo.SetFromFilename(c_file.name);
		if (!demoinfo.valid || !demoinfo.timeIsSet || demoinfo.cpm != processCPM || strcmp(demoinfo.playerName,gPlayerName))
			continue;
		time = demoinfo.GetTime();
		if (time == -1)
			continue;

		//sort time with previously processed demos
		i = 0;
		while (i < besttime_n && i < gNumDemosPerMap && besttime[i] < time)
			i++;

		if (i<=besttime_n && i<gNumDemosPerMap) {
			if (besttime_n < gNumDemosPerMap)
				besttime_n++;
			for(j=besttime_n-1; j>i; j--) {
				besttime[j] = besttime[j-1];
			}
			besttime[i] = time;
		}
	} while (_findnext(hFile,&c_file) == 0);

	_findclose(hFile);
	maxtime = besttime[besttime_n-1];


	// Second pass: delete demos whose time is above maxtime
	nDemosKept = 0;
	NoMoreFile = 0;
	hFile = _findfirst(va("%s%s%s*_%s_*%s",DEMO_PATH,mapname,(processCPM)?"_cpm_":"",gPlayerName,DEMO_EXT),&c_file);
	while (hFile != -1L && !NoMoreFile) {

		FileOpFailed = 0;

		//get time from demo filename
		demoinfo.SetFromFilename(c_file.name);

		if (demoinfo.valid && demoinfo.cpm != processCPM) {
			NoMoreFile = _findnext(hFile,&c_file);
			continue;
		}

		if (!demoinfo.valid || !demoinfo.timeIsSet || strcmp(demoinfo.playerName,gPlayerName)) {
			if (RenameToTmp(va("%s%s",DEMO_PATH,c_file.name))) {
				_findclose(hFile);
				hFile = _findfirst(va("%s%s*%s*%s",DEMO_PATH,mapname,gPlayerName,DEMO_EXT),&c_file);
			} else {
				NoMoreFile = _findnext(hFile,&c_file);
			}
			continue;
		}

		time = demoinfo.GetTime();

		//Check if demo should be deleted.
		//Rename its extension to .tmp otherwise so we don't process it again.
		if (time == -1 || (time <= maxtime && nDemosKept < gNumDemosPerMap)) {
			FileOpFailed = (RenameToTmp(va("%s%s",DEMO_PATH,c_file.name))) ? 0 : 1;
			if (time != -1)
				nDemosKept++;
		} else {
			if (FileOpFailed = remove(va("%s%s",DEMO_PATH,c_file.name)) == 0) {
				gStatsFilesNum++;
				gStatsSize += c_file.size / 1024;
			} else {
				trap_print(va("warning: can't delete %s.",c_file.name));
			}
		}

		if (FileOpFailed) {
			NoMoreFile = _findnext(hFile,&c_file);
		} else {
			hFile = _findfirst(va("%s%s*%s*%s",DEMO_PATH,mapname,gPlayerName,DEMO_EXT),&c_file);
		}

	}
	_findclose(hFile);
}

void DeleteExcessDemos() {
	struct _finddata_t c_file;
	long hFile;
	demoinfo_c demoinfo;
	int NoMoreFile;

	//process each demo from selected player, get map name, process each demo for that map, 
	//rename these to .tmp and go on
	gStatsFilesNum = 0;
	gStatsSize = 0;
	hFile = _findfirst(va("%s*%s*%s",DEMO_PATH,gPlayerName,DEMO_EXT),&c_file);
	NoMoreFile = 0;

	while (hFile != -1L && !NoMoreFile) {
		demoinfo.SetFromFilename(c_file.name);
		if (!demoinfo.valid || !demoinfo.timeIsSet) {
			if (RenameToTmp(c_file.name)) {
				_findclose(hFile);	//must release the search-handle and restart search after any file operation
				hFile = _findfirst(va("%s*%s*%s",DEMO_PATH,gPlayerName,DEMO_EXT),&c_file);
			} else {	//can't rename the demo
				NoMoreFile = _findnext(hFile,&c_file);	//will set NoMoreFile to 0 if a next file is found
			}
			continue;
		}
		ProcessDemosPerMap(demoinfo.mapName,demoinfo.cpm);

		if (fileExists(c_file.name)) {	//file still exists, something went wrong with .tmp renaming
			NoMoreFile = _findnext(hFile,&c_file);
		} else {
			_findclose(hFile);
			hFile = _findfirst(va("%s*%s*%s",DEMO_PATH,gPlayerName,DEMO_EXT),&c_file);
		}
	}
	_findclose(hFile);

	//rename .tmp'ed demos
	hFile = _findfirst(va("%s*%s*.tmp",DEMO_PATH,gPlayerName),&c_file);
	NoMoreFile = 0;

	while (hFile != -1L && !NoMoreFile) {
		if (RenameTmpToDemoExt(va("%s%s",DEMO_PATH,c_file.name))) {
			_findclose(hFile);	//must release the search-handle and restart search after renaming operation
			hFile = _findfirst(va("%s*%s*.tmp",DEMO_PATH,gPlayerName),&c_file);
		} else {
			NoMoreFile = _findnext(hFile,&c_file);
		}
	}
	_findclose(hFile);

	trap_print(va("Deleted %i files. (%i kb)\r\n",gStatsFilesNum,gStatsSize));
}


// ==============================================
//
//   Engine entry func
//
// ==============================================


//must be called from OS interface
void DT_run() {

	if (gNumDemosPerMap < 1)
		gNumDemosPerMap = 1;
	else if (gNumDemosPerMap > 16)
		gNumDemosPerMap = 16;

	//Pre-steps
	ConvertDM66Demos();
	ConvertDM67Demos();
	SanitizeFilenames();

	//Step 1 : Search and perform all .dem files
	trap_print("> Step 1\r\nRenaming demos from .dem files...\r\n");
	RenameRegisteredDemos();
	trap_print("\r\n");

	//Step 2 : Search and delete any demo that contains the "TimeHere" tag
	trap_print("> Step 2\r\nDeleting demos from unfinished runs...\r\n");
	DeleteUnfinishedDemos();
	trap_print("\r\n");

	//Step 3 : Keep only the nDemosPerMap demos made by the designed player for each map
	trap_print("> Step 3\r\nDeleting excess demos for each map...\r\n");
	DeleteExcessDemos();
	trap_print("\r\n");
}



